import keras_quadopt as kqp
import pytest
import tensorflow as tf


def test1():
    with pytest.raises(Exception):
        kqp.get_weights()


def test2():
    with pytest.raises(Exception):
        kqp.get_weights([1])


def test3():
    with pytest.raises(Exception):
        kqp.get_weights([1], [1], lam=-1)
    with pytest.raises(Exception):
        kqp.get_weights([1], [1], bnd_up=-1)
    with pytest.raises(Exception):
        kqp.get_weights([1], [1], bnd_up=2)


def test4():
    good = [.51, .52]
    simi = [[1., 2], [3, 4]]
    weights, results = kqp.get_weights(good, simi, lam=0.5)
    assert len(weights) == 2


def test5():
    # goodness scores
    good = tf.constant([.51, .53, .55, .57])
    # similarity matrices
    simi_1 = tf.constant([
        [1, .9, .8, .7],
        [.9, 1, .6, .5],
        [.8, .6, 1, .4],
        [.7, .5, .4, 1],
    ])
    simi_2 = tf.constant([
        [1, .7, .8, .3],
        [.7, 1, .4, .2],
        [.8, .4, 1, .6],
        [.3, .2, .6, 1],
    ])
    # preference parameters
    lam = 0.4
    beta_1 = 0.25
    beta_2 = 0.75
    # compute
    simi = kqp.aggregate_matrices(simi_1, beta_1, simi_2, beta_2)
    weights, _ = kqp.get_weights(good, simi, lam)
    # check
    assert simi.shape == (4, 4)
    assert weights.shape == (4,)



import keras_quadopt as kqp
import pytest
import tensorflow as tf


def test1():
    with pytest.raises(Exception):
        kqp.aggregate_matrices()


def test2():
    with pytest.raises(Exception):
        kqp.aggregate_matrices([1])


def test3():
    with pytest.raises(Exception):
        kqp.aggregate_matrices([1], 1, [2])


def test4():
    simi1 = [[1., 2], [3, 4]]
    beta1 = 2.
    simi2 = [[5., 6], [7, 8]]
    beta2 = 0.5
    simi = kqp.aggregate_matrices(simi1, beta1, simi2, beta2)
    tf.debugging.assert_equal(simi, [[4.5, 7], [9.5, 12]])


def test5():
    simi1 = [[1, 2], [3, 4]]
    beta1 = 2
    simi2 = [[5, 6], [7, 8]]
    beta2 = 0.5
    simi = kqp.aggregate_matrices(simi1, beta1, simi2, beta2, dtype=tf.float16)
    assert simi.dtype == tf.float16
    tf.debugging.assert_equal(simi, tf.constant(
        [[4.5, 7], [9.5, 12]], dtype=tf.float16))



import tensorflow as tf


def aggregate_matrices(*args, dtype=tf.float32):
    # check inputs
    if len(args) % 2:
        raise Exception("An even number of input arguments expected")
    # copy the first matrix
    out = tf.constant(args[0], dtype=dtype) * tf.constant(args[1], dtype=dtype)
    # add the other matrices
    for i in range(2, len(args), 2):
        out += (tf.constant(args[i], dtype=dtype) * tf.constant(
            args[i + 1], dtype=dtype))
    # done
    return out



import tensorflow as tf
from typing import List, Optional


def get_weights(c: List[float],
                Q: List[List[float]],
                lam: Optional[float] = 0.5,
                dtype=tf.float32,
                maxiter: Optional[int] = 500,
                ftol: Optional[float] = 1e-06,
                patience: int = 50
                ):
    """Maximize "total goodness" and minimize "total similarity" """
    # error checking
    if lam < 0:
        raise Exception(f"the preference lambda='{lam}' must be positive")

    # how big is `N`
    n_examples = len(c)

    # We can multipy `lam*Q` beforehand and save compute time!
    lamQ = tf.constant(Q) * lam  # (n,n)
    cin = tf.constant(c)  # (n,)

    # initial values
    w0 = tf.ones(shape=(n_examples,)) / n_examples
    # trainable params
    w = tf.Variable(
        initial_value=w0, trainable=True, dtype=dtype)

    # loss function with regularization
    def custom_loss():
        # norm to 1
        v = w / tf.maximum(1e-8, tf.reduce_sum(w))
        # quadratic problem
        loss = -tf.tensordot(cin, v, axes=1)
        loss += tf.sqrt(tf.tensordot(tf.tensordot(lamQ, v, axes=1), v, axes=1))
        # regularization: sum_i w_i = 1
        loss += tf.pow(1. - tf.reduce_sum(w), 2)
        # regularization: w_i >= 0
        loss += tf.reduce_sum(-tf.minimum(w, 0.0))
        return loss

    # optimizer = tf.keras.optimizers.SGD(learning_rate=0.1)
    optimizer = tf.keras.optimizers.Adam(
        learning_rate=0.1, beta_1=.9, beta_2=.999,
        epsilon=1e-7, amsgrad=True)  # 3e-4

    # start loop
    fbest = custom_loss()
    wbest = w.numpy()
    wait = 0
    for i in range(maxiter):
        optimizer.minimize(loss=custom_loss, var_list=[w])
        f = custom_loss()
        if fbest > (f + ftol):
            fbest = f
            wbest = w.numpy()
            wait = 0
        else:
            wait += 1
            if wait > patience:
                break

    # done
    return wbest, fbest
